// semaphore_manager.h
#ifndef SEMAPHORE_MANAGER_H
#define SEMAPHORE_MANAGER_H

#include <systemc>
#include <vector>
#include <queue>
#include <iostream>
#include <mutex>
#include <unordered_set>
#include "configuration.h"

using namespace sc_core;

class semaphore_manager : public sc_module {
public:
    SC_HAS_PROCESS(semaphore_manager);

    semaphore_manager(sc_module_name name) : sc_module(name) {
        semaphore_owners.resize(ROUTER_NO, 0);      // Start with all tokens at router 0
        request_queues.resize(ROUTER_NO);           // Initialize a queue for each semaphore
        queue_entries.resize(ROUTER_NO);                // Set for each semaphore to track queued routers
        semaphore_status.resize(ROUTER_NO, false); // All semaphores are initially free
        std::fill_n(last_router_owner, ROUTER_NO, -1); // Initialize all last owners to -1
        SC_THREAD(update_semaphores);               // Start automatic ownership update thread
        sc_report_handler::set_log_file_name("out/report.log");
        sc_report_handler::set_actions(LOG_NAME, SC_INFO, SC_LOG|SC_DISPLAY);
    }

    // Request a token for a semaphore with round-robin and priority ordering
    bool request_token(int router_id, int semaphore_id) {
        //wait(SC_ZERO_TIME);
        std::lock_guard<sc_mutex> lock(mutex);  // Ensure atomic access
        if (semaphore_id >= 0 && semaphore_id < ROUTER_NO) {
            
            wait(SC_ZERO_TIME);
            // Give priority if the request follows round-robin order from the last owner
            if (semaphore_owners[semaphore_id] == router_id || semaphore_owners[semaphore_id] == -1 || (!semaphore_status[semaphore_id] && request_queues[semaphore_id].empty())) 
                {
                log_info("Token for destination router " + std::to_string(semaphore_id) + " granted to Router " + std::to_string(router_id));
                semaphore_owners[semaphore_id] = router_id;
                semaphore_status[semaphore_id] = true;  // Mark semaphore as in use
                last_router_owner[semaphore_id] = router_id; // Update the last owner
                return true;
            } else {
               if (queue_entries[semaphore_id].find(router_id) == queue_entries[semaphore_id].end()) {
                    request_queues[semaphore_id].push(router_id);
                    queue_entries[semaphore_id].insert(router_id);  // Track in the set
                    log_info("Router " + std::to_string(router_id) + " added to queue for semaphore " + std::to_string(semaphore_id));
                }
               
            }
        }
        return false;
    }

    // Release the token and pass it to the next in the round-robin or queue order
    bool release_token(int router_id, int semaphore_id) {
        std::lock_guard<sc_mutex> lock(mutex);  // Ensure atomic access
        if (semaphore_id >= 0 && semaphore_id < ROUTER_NO) {
            if (semaphore_owners[semaphore_id] == router_id) {
                log_info("Router " + std::to_string(router_id) + " released token for destination router " + std::to_string(semaphore_id));
                 semaphore_status[semaphore_id] = false; // Mark semaphore as free

                //wait(0, SC_NS);

                if (!request_queues[semaphore_id].empty()) {
                    int next_router = request_queues[semaphore_id].front();
                    request_queues[semaphore_id].pop();
                    queue_entries[semaphore_id].erase(next_router);  // Remove from set when dequeued
                    semaphore_owners[semaphore_id] = next_router;
                    semaphore_status[semaphore_id] = true;  // Mark semaphore as in use
                    log_info("Token for destination router " + std::to_string(semaphore_id) + " passed to Router " + std::to_string(next_router));
                    //request_token(next_router, semaphore_id);
                    
                } else {
                    semaphore_owners[semaphore_id] = -1;
                    // Assign to next router in round-robin if idle
                    // semaphore_owners[semaphore_id] = (semaphore_owners[semaphore_id] + 1) % ROUTER_NO;
                    // log_info("Token for destination router " + std::to_string(semaphore_id) + " passed to Router " + std::to_string(semaphore_owners[semaphore_id]));
                }
                return true;
            }
        }
        return false;
    }

    // Periodic update function to rotate token ownership in round-robin if tokens are idle
    void update_semaphores() {
        while (true) {
            wait(20, SC_NS);
            for (int i = 0; i < ROUTER_NO; ++i) {
                std::lock_guard<sc_mutex> lock(mutex);  // Ensure atomic update
                if ((!semaphore_status[i] && (semaphore_owners[i] == -1 || request_queues[i].empty()))) {
                    semaphore_owners[i] = (last_router_owner[i] + 1) % ROUTER_NO;
                    log_info("Updated owner of semaphore " + std::to_string(i) + " to Router " + std::to_string(semaphore_owners[i]));
                    last_router_owner[i] = semaphore_owners[i];
                    
                }
            }
        }
    }

    
private:
    std::vector<int> semaphore_owners;                // Tracks router ownership (-1 means no owner)
    std::vector<std::queue<int>> request_queues;      // Queue of pending requests per semaphore
    std::vector<bool> semaphore_status;               // Semaphore availability
    sc_mutex mutex;                                   // Mutex to ensure atomic operations
    std::vector<std::unordered_set<int>> queue_entries; // Track entries in each queue to prevent duplicates
    int last_router_owner[ROUTER_NO];  // Array to track last owner for round-robin assignment

    //print log messages
    void log_info(std::string msg){
        SC_REPORT_INFO(LOG_NAME, msg.c_str());
    }

    void log_error(std::string msg){
        SC_REPORT_ERROR(LOG_NAME, msg.c_str());
    }

};

#endif // SEMAPHORE_MANAGER_H
