#include "core.h"

// Initiator module (processor core)
core::core(sc_module_name name) {
    sc_report_handler::set_log_file_name("out/report.log");
    sc_report_handler::set_actions(LOG_NAME, SC_INFO, SC_LOG|SC_DISPLAY);
    SC_THREAD(thread);
    target_socket.register_nb_transport_fw(this, &core::nb_transport_fw);
}

// Thread for sending data
void core::thread() {

 // Initialize random seed (only once per run, for all cores)
    static bool seeded = false;
    if (!seeded) {
        srand(time(0));  // Seed the random number generator
        seeded = true;
    }

    // Infinite loop for sending data
    while (true) {
        my_payload* trans = nullptr;
        // Core ID and source router ID from main logic
        unsigned int source_core = source_core_id;   
        unsigned int src_router_id = source_router_id;  // Source router from the main logic

        sc_time delay = SC_ZERO_TIME;  // Start with zero delay
        sc_time inter_transaction_delay = sc_time(90, SC_NS);  // Delay between transactions
        // Loop to retry the same transaction
        bool retry = false;
        do {
            if (!retry) {
        // Generate random data between 0 and 100
        int data = rand() % 101;

        // Random destination router (ensuring it's different from source router)
        unsigned int destination_router_id;
        do {
            destination_router_id = rand() % ROUTER_NO;  // Random router ID between 0 and ROUTER_NO-1
        } while (destination_router_id == src_router_id);  // Ensure it's not the same as source router

        // Destination core ID (random between 0 and CORE_NO-1)
        unsigned int dst_core_id = rand() % CORE_NO;

        // Create payload with generated values
         trans = new my_payload(source_core, src_router_id, dst_core_id, destination_router_id, data);
         trans->update_timestamp(sc_time_stamp());  // Set the current simulation time

        // Construct the address (calculation based on router/core IDs)
        unsigned int offset = 0x20;  // offset
        unsigned int address = (destination_router_id << 12) | (dst_core_id << 8) | offset;
        trans->set_address(address);

        // Print which core is currently sending the data
        log_info( "Core " + std::to_string(source_core) + " of Source Router: " + std::to_string(source_router_id) 
                + ": Processing data " + std::to_string(trans->data) + " for router " + std::to_string(destination_router_id) 
                + " and core " + std::to_string(dst_core_id) + " at time: " + trans->timestamp.to_string());
            }
        
        // Set up the rest of the transaction
        trans->set_byte_enable_ptr(nullptr);
        trans->set_dmi_allowed(false);
        trans->set_response_status(tlm::TLM_INCOMPLETE_RESPONSE);
        
        
        tlm::tlm_phase phase = tlm::BEGIN_REQ; // Transaction phase
        tlm::tlm_sync_enum status = socket->nb_transport_fw(*trans, phase, delay); // Non-blocking transport

        if (status == tlm::TLM_COMPLETED || phase == tlm::END_REQ) {
            log_info("Core " + std::string(name()) + ": Transaction successful, data sent: " + std::to_string(trans->data) + " at time: " + sc_time_stamp().to_string());
            delete trans; // Clean up the transaction
            retry = false;  // Move on to the next transaction
            wait(inter_transaction_delay);
        } else if (status == tlm::TLM_ACCEPTED) {
            log_info("Core " + std::string(name()) + ": Transaction accepted, waiting for completion.");
            wait(delay);  // Wait for response
            retry = false;  // No need to retry, accepted transaction will continue
        } else if (status == tlm::TLM_UPDATED && phase == tlm::BEGIN_REQ) {
            log_info("Core " + std::string(name()) + ": FIFO full, retrying...");
            //wait(delay); // Wait for some time before retrying
           // --i; // Retry the same transaction
           wait(delay + sc_time(20, SC_NS));  // Add some delay before retrying
         // wait(router->core_fifo_event);
            retry = true;  // Retry the same transaction
        }
    
    } while (retry);

    
    wait(delay);
    }
}

/// Forwarding function for receiving data from router
tlm::tlm_sync_enum core::nb_transport_fw(tlm::tlm_generic_payload& trans, tlm::tlm_phase& phase, sc_time& delay) {
    if (phase == tlm::BEGIN_REQ) {
        my_payload* my_trans = dynamic_cast<my_payload*>(&trans);

    // Print which core is processing the received data
        log_info("Core " + std::to_string(source_core_id) + " of parent Router: " + std::to_string(source_router_id) 
                + ": Processing  transaction " + "and received data: " + std::to_string(my_trans->data));

       // log_info("Core " + std::string(name()) + ": Received transaction, data: " + std::to_string(my_trans->data));
        
        
        // Send acknowledgment or status back 
        trans.set_response_status(tlm::TLM_OK_RESPONSE);

        // Log the received data and remove the payload
        delete my_trans;
        return tlm::TLM_COMPLETED;
    }
    return tlm::TLM_ACCEPTED;
}



void core::log_info(std::string msg){
    SC_REPORT_INFO(LOG_NAME, msg.c_str());
}

void core::log_error(std::string msg){
    SC_REPORT_ERROR(LOG_NAME, msg.c_str());
}